;; Generated package description from compat.el  -*- no-byte-compile: t -*-
(define-package "compat" "28.1.2.2.0.20221122.225006" "Emacs Lisp Compatibility Library" '((emacs "24.3") (nadvice "0.3")) :commit "1573aa2e6d478c30cc7f394b72cbaa4db77ef811" :authors '(("Philip Kaludercic" . "philipk@posteo.net")) :maintainer '("Compat Development" . "~pkal/compat-devel@lists.sr.ht") :keywords '("lisp") :url "https://sr.ht/~pkal/compat")
